unsigned long int* PIO_PERC = (unsigned long int*) 0xfffff800;
unsigned long int* PIO_OERC = (unsigned long int*) 0xfffff810;
unsigned long int* PIO_ODRC = (unsigned long int*) 0xfffff814;
unsigned long int* PIO_SODRC = (unsigned long int*) 0xfffff830;
unsigned long int* PIO_CODRC = (unsigned long int*) 0xfffff834;
unsigned long int* PIO_PDSRC = (unsigned long int*) 0xfffff83c;

/* Mapping from keypad matrix column line to GPIO pin */
const unsigned char column_list[]=
{
   0,    /* Col 0 is pin PC0 */
   1,    /* Col 1 is pin PC1 */
   2,    /* Col 2 is pin PC2 */
   3,    /* Col 3 is pin PC3 */
   4,    /* Col 4 is pin PC4 */
   5,    /* Col 5 is pin PC5 */
   6     /* Col 6 is pin PC6 */
};

/* Mapping from keypad matrix row line to GPIO pin */
const unsigned char row_list[]=
{
   11,      /* Row 0 is pin PC11 */
   12,      /* Row 1 is pin PC12 */
   13,      /* Row 2 is pin PC13 */
   14,      /* Row 3 is pin PC14 */
   15,      /* Row 4 is pin PC15 */
   26       /* Row 5 is pin PC26 */
};

/* Set GPIO to work with the matrix keypad */
/* We'll plan on driving the columns and reading the rows */
int matrix_init()
{
   int i;

   /* Configure the column lines as outputs and idle them high */
   for (i=0; i<7; i++)
   {
      *(unsigned long int*)PIO_PERC = (1 << column_list[i]);      /* Assign pin to GPIO */
      *(unsigned long int*)PIO_OERC = (1 << column_list[i]);      /* Make it an output */
      *(unsigned long int*)PIO_CODRC = (1 << column_list[i]);      /* Drive it high */
   }

   /* Configure the row lines as inputs */
   for (i=0; i<6; i++)
   {
      *(unsigned long int*)PIO_PERC = (1 << row_list[i]);      /* Assign pin as GPIO */
      *(unsigned long int*)PIO_ODRC = (1 << row_list[i]);      /* And make each row an input */
   }

   /* On switch is a special case--it's not in the matrix, but tied to PC10 */
   /* We'll make it an input  */
   *(unsigned long int*)PIO_PERC = (1 << 10);      /* Assign to GPIO */
   *(unsigned long int*)PIO_ODRC = (1 << 10);      /* And make that pin an input pin */

   return 0;
}

/* Scan the matrix and return a scancode if a key is pressed */
/* If no key is pressed, return -1 */
/* Scan code is one byte, high nibble the column, low nibble the row */
/* Rows and columns are zero based */
/* First key detected ends the scan */
int key_scan()
{
   int column;
   int row;
   unsigned long int input;

   /* Check for the on key first (not in the matrix */
   input=*(unsigned long int*)PIO_PDSRC;
   if ((~input) & (1 << 10))
   {
      /* Make up the scancode for where the on key would be, row 6 column 0 */
      return 0x60;
   }

   /* Take each column low in turn */
   for (column=0; column<7; column++)
   {
      /* drive a column low */
      *(unsigned long int*)PIO_CODRC = (1 << column_list[column]);

      /* Grab the input lines */
      input=*(unsigned long int*)PIO_PDSRC;

      /* Restore the column line to idle high condition */
      *(unsigned long int*)PIO_SODRC = (1 << column_list[column]);

      /* Examine the row and see if any keys were pressed */
      /* Pressed keys will pull the row pins low */
      for (row=0; row<6; row++)
      {
         /* Is the row line being pulled low? */
         if ((~input) & (1 << row_list[row]))
         {
            /* Yes, so it was pressed. */
            /* Generate the scancode from the row and column */
            return (column << 4) | (row);
         }
      }
   }

   /* No keys pressed, so return -1 to indicate it */
   return -1;
}


/* Font table for digits 0-9 and A-F */
/* The segment mapping must match the big LCD lookup table below */
/* Yes, it's an ugly hack, but it does work. */
/* Feel free to do a better one! */
const unsigned char seven_segment_table[]=
{
   0x7b,    /* 0 */
   0x48,    /* 1 */
   0x3d,    /* 2 */
   0x6d,    /* 3 */
   0x4e,    /* 4 */
   0x67,    /* 5 */
   0x77,    /* 6 */
   0x49,    /* 7 */
   0x7f,    /* 8 */
   0x6f,    /* 9 */
   0x5f,    /* A */
   0x76,    /* B */
   0x34,    /* C */
   0x7c,    /* D */
   0x37,    /* E */
   0x17,    /* F */
};


struct lcd_digit_struct
{
   unsigned long int address;
   unsigned long int bitmask;
};

/* Table of LCD segments */
/* Eight entries for each of the 12 main digits (left to right) */
/* This table includes segments a-h, but not i, the comma */
/* Again, this is an ugly hack and needs to be cleaned up */
/* I envision a table of relative offsets and bit numbers, easily 1/4 the size */
const struct lcd_digit_struct lcd_digit_table[]=
{
   {0xfffb4210, 0x80},
   {0xfffb4210, 0x40},
   {0xfffb4218, 0x80},
   {0xfffb4210, 0x100},
   {0xfffb4218, 0x40},
   {0xfffb4220, 0x40},
   {0xfffb4220, 0x80},
   {0xfffb4228, 0x100},

   {0xfffb4210, 0x200},
   {0xfffb4218, 0x100},
   {0xfffb4218, 0x200},
   {0xfffb4218, 0x400},
   {0xfffb4220, 0x100},
   {0xfffb4220, 0x200},
   {0xfffb4220, 0x400},
   {0xfffb4228, 0x400},

   {0xfffb4210, 0x800},
   {0xfffb4210, 0x400},
   {0xfffb4218, 0x800},
   {0xfffb4210, 0x1000},
   {0xfffb4220, 0x800},
   {0xfffb4228, 0x800},
   {0xfffb4218, 0x1000},
   {0xfffb4220, 0x1000},

   {0xfffb4210, 0x4000},
   {0xfffb4210, 0x2000},
   {0xfffb4218, 0x2000},
   {0xfffb4218, 0x4000},
   {0xfffb4220, 0x2000},
   {0xfffb4228, 0x2000},
   {0xfffb4220, 0x4000},
   {0xfffb4220, 0x8000},

   {0xfffb4210, 0x10000},
   {0xfffb4210, 0x8000},
   {0xfffb4218, 0x10000},
   {0xfffb4210, 0x20000},
   {0xfffb4218, 0x8000},
   {0xfffb4228, 0x8000},
   {0xfffb4220, 0x10000},
   {0xfffb4228, 0x20000},

   {0xfffb4210, 0x40000},
   {0xfffb4218, 0x20000},
   {0xfffb4218, 0x40000},
   {0xfffb4218, 0x80000},
   {0xfffb4220, 0x20000},
   {0xfffb4220, 0x40000},
   {0xfffb4220, 0x80000},
   {0xfffb4228, 0x80000},

   {0xfffb4210, 0x100000},
   {0xfffb4210, 0x80000},
   {0xfffb4218, 0x100000},
   {0xfffb4210, 0x200000},
   {0xfffb4220, 0x100000},
   {0xfffb4228, 0x100000},
   {0xfffb4218, 0x200000},
   {0xfffb4220, 0x200000},

   {0xfffb4210, 0x800000},
   {0xfffb4210, 0x400000},
   {0xfffb4218, 0x400000},
   {0xfffb4218, 0x800000},
   {0xfffb4220, 0x400000},
   {0xfffb4228, 0x400000},
   {0xfffb4220, 0x800000},
   {0xfffb4220, 0x1000000},

   {0xfffb4210, 0x2000000},
   {0xfffb4210, 0x1000000},
   {0xfffb4218, 0x2000000},
   {0xfffb4210, 0x4000000},
   {0xfffb4218, 0x1000000},
   {0xfffb4228, 0x1000000},
   {0xfffb4220, 0x2000000},
   {0xfffb4228, 0x4000000},

   {0xfffb4210, 0x8000000},
   {0xfffb4218, 0x4000000},
   {0xfffb4218, 0x8000000},
   {0xfffb4218, 0x10000000},
   {0xfffb4220, 0x4000000},
   {0xfffb4220, 0x8000000},
   {0xfffb4220, 0x10000000},
   {0xfffb4228, 0x10000000},

   {0xfffb4210, 0x20000000},
   {0xfffb4210, 0x10000000},
   {0xfffb4218, 0x20000000},
   {0xfffb4210, 0x40000000},
   {0xfffb4220, 0x20000000},
   {0xfffb4228, 0x20000000},
   {0xfffb4218, 0x40000000},
   {0xfffb4220, 0x40000000},

   {0xfffb4214, 0x1},
   {0xfffb4210, 0x80000000},
   {0xfffb4218, 0x80000000},
   {0xfffb421c, 0x1},
   {0xfffb4220, 0x80000000},
   {0xfffb4228, 0x80000000},
   {0xfffb4224, 0x1},
   {0xfffb422c, 0x2},
};

/* Place a hex digit (value) at position (digit) on the main 12 digit display */
/* Values > 15 will show as a blank */
void write_char(int digit, int value)
{
   int pattern;
   int bit;
   unsigned long int* address;

   /* Font lookup--convert to segments */
   pattern=seven_segment_table[value];

   /* Blank if out of range */
   if (value>15)
      pattern=0;

   /* Look up each segment in the main LCD table and turn 'em on or off */
   for (bit=0; bit<8; bit++)
   {
      /* Grab address of the segment */
      address=(unsigned long int*)lcd_digit_table[(digit*8)+bit].address;

      /* Does font call for segment to be on? */
      if (pattern & (1 << bit))
      {
         /* Yep, so look up the bit and set it */
         *address|=lcd_digit_table[(digit*8)+bit].bitmask;
      }
      else
      {
         /* Segment should be off; lookup and turn if off */
         *address&=~(lcd_digit_table[(digit*8)+bit].bitmask);
      }
   }
}

/* Scan keyboard */
/* If a key is pressed, display the scancode */
/* Column is shown in digit 0 (leftmost), row in digit 1 */
/* If no key is pressed, display is blank */
void test_digits()
{
   int digit=0;

   /* Scan matrix */
   digit=key_scan();

   if (digit>=0)
   {
      /* Valid scancode, so something is pressed */
      /* Show column */
      write_char(0, (digit & 0xf0)>>4);

      /* Show row */
      write_char(1, digit & 0x0f);
   }
   else
   {
      /* No key pressed, blank display */
      write_char(0, 16);
      write_char(1, 16);
   }
}

/* Turn off all the LCD segments */
void clear_screen()
{
   unsigned long int* i;
   for (i=(unsigned long int*)0xfffb4200; i<(unsigned long int*)0xfffb4250; i++)
   {
      *i=0;
   }
}

int main(void)
{
   /* Setup GPIO to scan keypad */
   matrix_init();

   /* Clear screen */
   clear_screen();

   /* Loop... */
   for (;;)
   {
      /* Show scancode */
      test_digits();
   }

   return 0;
}
